package com.simibubi.create.content.processing.recipe;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import com.simibubi.create.Create;
import com.simibubi.create.foundation.data.SimpleDatagenIngredient;
import com.simibubi.create.foundation.data.recipe.Mods;
import com.simibubi.create.foundation.fluid.FluidHelper;
import com.simibubi.create.foundation.fluid.FluidIngredient;
import com.simibubi.create.foundation.recipe.IRecipeTypeInfo;
import com.simibubi.create.foundation.utility.Pair;
import com.tterrag.registrate.util.DataIngredient;

import io.github.fabricators_of_create.porting_lib.fluids.FluidStack;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredient;
import net.fabricmc.fabric.api.resource.conditions.v1.ConditionJsonProvider;
import net.fabricmc.fabric.api.resource.conditions.v1.DefaultResourceConditions;
import net.fabricmc.fabric.api.resource.conditions.v1.ResourceConditions;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1856;
import net.minecraft.class_1865;
import net.minecraft.class_1935;
import net.minecraft.class_2371;
import net.minecraft.class_2444;
import net.minecraft.class_2960;
import net.minecraft.class_3611;
import net.minecraft.class_6862;

public class ProcessingRecipeBuilder<T extends ProcessingRecipe<?>> {

	protected ProcessingRecipeFactory<T> factory;
	protected ProcessingRecipeParams params;
	protected List<ConditionJsonProvider> recipeConditions;

	public ProcessingRecipeBuilder(ProcessingRecipeFactory<T> factory, class_2960 recipeId) {
		params = new ProcessingRecipeParams(recipeId);
		recipeConditions = new ArrayList<>();
		this.factory = factory;
	}

	public ProcessingRecipeBuilder<T> withItemIngredients(class_1856... ingredients) {
		return withItemIngredients(class_2371.method_10212(class_1856.field_9017, ingredients));
	}

	public ProcessingRecipeBuilder<T> withItemIngredients(class_2371<class_1856> ingredients) {
		params.ingredients = ingredients;
		return this;
	}

	public ProcessingRecipeBuilder<T> withSingleItemOutput(class_1799 output) {
		return withItemOutputs(new ProcessingOutput(output, 1));
	}

	public ProcessingRecipeBuilder<T> withItemOutputs(ProcessingOutput... outputs) {
		return withItemOutputs(class_2371.method_10212(ProcessingOutput.EMPTY, outputs));
	}

	public ProcessingRecipeBuilder<T> withItemOutputs(class_2371<ProcessingOutput> outputs) {
		params.results = outputs;
		return this;
	}

	public ProcessingRecipeBuilder<T> withFluidIngredients(FluidIngredient... ingredients) {
		return withFluidIngredients(class_2371.method_10212(FluidIngredient.EMPTY, ingredients));
	}

	public ProcessingRecipeBuilder<T> withFluidIngredients(class_2371<FluidIngredient> ingredients) {
		params.fluidIngredients = ingredients;
		return this;
	}

	public ProcessingRecipeBuilder<T> withFluidOutputs(FluidStack... outputs) {
		return withFluidOutputs(class_2371.method_10212(FluidStack.EMPTY, outputs));
	}

	public ProcessingRecipeBuilder<T> withFluidOutputs(class_2371<FluidStack> outputs) {
		params.fluidResults = outputs;
		return this;
	}

	public ProcessingRecipeBuilder<T> duration(int ticks) {
		params.processingDuration = ticks;
		return this;
	}

	public ProcessingRecipeBuilder<T> averageProcessingDuration() {
		return duration(100);
	}

	public ProcessingRecipeBuilder<T> requiresHeat(HeatCondition condition) {
		params.requiredHeat = condition;
		return this;
	}

	public T build() {
		validateFluidAmounts();
		return factory.create(params);
	}

	public void build(Consumer<class_2444> consumer) {
		consumer.accept(new DataGenResult<>(build(), recipeConditions));
	}

	public static final long[] SUS_AMOUNTS = { 10, 250, 500, 1000 };

	private void validateFluidAmounts() {
		for (FluidIngredient ingredient : params.fluidIngredients) {
			for (long amount : SUS_AMOUNTS) {
				if (ingredient.getRequiredAmount() == amount) {
					Create.LOGGER.warn("Suspicious fluid amount in recipe [{}]: {}", params.id, amount);
				}
			}
		}
	}

	// Datagen shortcuts

	public ProcessingRecipeBuilder<T> require(class_6862<class_1792> tag) {
		return require(class_1856.method_8106(tag));
	}

	public ProcessingRecipeBuilder<T> require(class_1935 item) {
		return require(class_1856.method_8091(item));
	}

	public ProcessingRecipeBuilder<T> require(class_1856 ingredient) {
		params.ingredients.add(ingredient);
		return this;
	}

	// fabric: custom ingredient support
	public ProcessingRecipeBuilder<T> require(CustomIngredient ingredient) {
		return require(ingredient.toVanilla());
	}

	public ProcessingRecipeBuilder<T> require(Mods mod, String id) {
		params.ingredients.add(new SimpleDatagenIngredient(mod, id));
		return this;
	}

	public ProcessingRecipeBuilder<T> require(class_2960 ingredient) {
		params.ingredients.add(DataIngredient.ingredient(null, ingredient));
		return this;
	}

	public ProcessingRecipeBuilder<T> require(class_3611 fluid, long amount) {
		return require(FluidIngredient.fromFluid(fluid, amount));
	}

	public ProcessingRecipeBuilder<T> require(class_6862<class_3611> fluidTag, long amount) {
		return require(FluidIngredient.fromTag(fluidTag, amount));
	}

	public ProcessingRecipeBuilder<T> require(FluidIngredient ingredient) {
		params.fluidIngredients.add(ingredient);
		return this;
	}

	public ProcessingRecipeBuilder<T> output(class_1935 item) {
		return output(item, 1);
	}

	public ProcessingRecipeBuilder<T> output(float chance, class_1935 item) {
		return output(chance, item, 1);
	}

	public ProcessingRecipeBuilder<T> output(class_1935 item, int amount) {
		return output(1, item, amount);
	}

	public ProcessingRecipeBuilder<T> output(float chance, class_1935 item, int amount) {
		return output(chance, new class_1799(item, amount));
	}

	public ProcessingRecipeBuilder<T> output(class_1799 output) {
		return output(1, output);
	}

	public ProcessingRecipeBuilder<T> output(float chance, class_1799 output) {
		return output(new ProcessingOutput(output, chance));
	}

	public ProcessingRecipeBuilder<T> output(float chance, Mods mod, String id, int amount) {
		return output(new ProcessingOutput(Pair.of(mod.asResource(id), amount), chance));
	}

	public ProcessingRecipeBuilder<T> output(float chance, class_2960 registryName, int amount) {
		return output(new ProcessingOutput(Pair.of(registryName, amount), chance));
	}

	public ProcessingRecipeBuilder<T> output(ProcessingOutput output) {
		params.results.add(output);
		return this;
	}

	public ProcessingRecipeBuilder<T> output(class_3611 fluid, long amount) {
		fluid = FluidHelper.convertToStill(fluid);
		return output(new FluidStack(fluid, amount));
	}

	public ProcessingRecipeBuilder<T> output(FluidStack fluidStack) {
		params.fluidResults.add(fluidStack);
		return this;
	}

	public ProcessingRecipeBuilder<T> toolNotConsumed() {
		params.keepHeldItem = true;
		return this;
	}

	//

	public ProcessingRecipeBuilder<T> whenModLoaded(String modid) {
		return withCondition(DefaultResourceConditions.allModsLoaded(modid));
	}

	public ProcessingRecipeBuilder<T> whenModMissing(String modid) {
		return withCondition(DefaultResourceConditions.not(DefaultResourceConditions.allModsLoaded(modid)));
	}

	public ProcessingRecipeBuilder<T> withCondition(ConditionJsonProvider condition) {
		recipeConditions.add(condition);
		return this;
	}

	@FunctionalInterface
	public interface ProcessingRecipeFactory<T extends ProcessingRecipe<?>> {
		T create(ProcessingRecipeParams params);
	}

	public static class ProcessingRecipeParams {

		protected class_2960 id;
		protected class_2371<class_1856> ingredients;
		protected class_2371<ProcessingOutput> results;
		protected class_2371<FluidIngredient> fluidIngredients;
		protected class_2371<FluidStack> fluidResults;
		protected int processingDuration;
		protected HeatCondition requiredHeat;

		public boolean keepHeldItem;

		protected ProcessingRecipeParams(class_2960 id) {
			this.id = id;
			ingredients = class_2371.method_10211();
			results = class_2371.method_10211();
			fluidIngredients = class_2371.method_10211();
			fluidResults = class_2371.method_10211();
			processingDuration = 0;
			requiredHeat = HeatCondition.NONE;
			keepHeldItem = false;
		}

	}

	public static class DataGenResult<S extends ProcessingRecipe<?>> implements class_2444 {

		private List<ConditionJsonProvider> recipeConditions;
		private ProcessingRecipeSerializer<S> serializer;
		private class_2960 id;
		private S recipe;

		@SuppressWarnings("unchecked")
		public DataGenResult(S recipe, List<ConditionJsonProvider> recipeConditions) {
			this.recipe = recipe;
			this.recipeConditions = recipeConditions;
			IRecipeTypeInfo recipeType = this.recipe.getTypeInfo();
			class_2960 typeId = recipeType.getId();

			if (!(recipeType.getSerializer() instanceof ProcessingRecipeSerializer))
				throw new IllegalStateException("Cannot datagen ProcessingRecipe of type: " + typeId);

			this.id = new class_2960(recipe.method_8114().method_12836(),
					typeId.method_12832() + "/" + recipe.method_8114().method_12832());
			this.serializer = (ProcessingRecipeSerializer<S>) recipe.method_8119();
		}

		@Override
		public void method_10416(JsonObject json) {
			serializer.write(json, recipe);
			if (recipeConditions.isEmpty())
				return;

			JsonArray conds = new JsonArray();
			recipeConditions.forEach(c -> conds.add(c.toJson())); // FabricDataGenHelper.addConditions(json, recipeConditions.toArray());?
			json.add(ResourceConditions.CONDITIONS_KEY, conds);
		}

		@Override
		public class_2960 method_10417() {
			return id;
		}

		@Override
		public class_1865<?> method_17800() {
			return serializer;
		}

		@Override
		public JsonObject method_10415() {
			return null;
		}

		@Override
		public class_2960 method_10418() {
			return null;
		}

	}

}
